# see: https://docs.python.org/3/library/socketserver.html
import logging
import socket
import socketserver
import traceback
from abc import abstractmethod, ABC
from decimal import Decimal
from typing import Tuple, Optional, Union, Any

from pandas import DataFrame

from pypsql_api.config.types import Session
from pypsql_api.ext.handlers import AuthHandler, QueryHandler, ExtendedQueryHandler
from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context
from pypsql_api.wire.back import ErrorResponse
from pypsql_api.wire.bytes import ReadingIO, WritingIO
from pypsql_api.wire.extended.types import Portal, PreparedStatement, Param
from pypsql_api.wire.front import BindMessage, ParseMessage
from pypsql_api.wire.protocol import run_states


def state_runner(existing_state, message_processor):
    state = message_processor.next(existing_state)
    message_processor.is_allowed(state)

    return state


class PsqlHandler(ABC, socketserver.StreamRequestHandler, socketserver.ThreadingMixIn):

    @property
    @abstractmethod
    def auth_handler_cls(self):
        raise Exception("Not implemented")

    @property
    @abstractmethod
    def socket_timeout(self):
        raise Exception("Not implemented")

    @property
    @abstractmethod
    def query_handler_cls(self):
        raise Exception("Not implemented")

    @abstractmethod
    def init_ssl(self, context: Context):
        raise Exception("Not implemented")

    def handle(self):

        print(f"First Connect")
        logging.info(f"First connect")

        # timeout causes the socket to become non blockgin
        # self.request.settimeout(self.socket_timeout())

        input = ReadingIO(self.rfile)
        output = WritingIO(self.wfile)

        try:
            context = Context(
                input=input,
                output=output,
                mem={},

                auth_handler=self.auth_handler_cls(),
                query_handler=self.query_handler_cls(),
                extended_query_handler=self.extended_query_handler_cls(),
                session=None,
                portals={},
                prepared_statements={},
                key_file=None,
                cert_file=None,
                socket=self.connection
            )

            self.init_ssl(context=context)

            state_name = Names.READ_SSL_REQUEST

            context, state_name = run_states(context, state_name)

        except socket.timeout as  e:
            print("connection timed out")
        except Exception as e:
            print(e)
            try:
                ErrorResponse.severe(str(e)).write(output)
            except:
                traceback.print_exc()
        finally:
            ssl_obj = context.mem.get('ssl_obj', None)
            if ssl_obj is None:
                output.close()
            else:
                ssl_obj.close()

        print("End")


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


def run(auth_handler_cls, query_handler_cls, extended_query_handler_cls, handler=None, host="", port=55432,
        cert_file=None, key_file=None, socket_timeout=240):
    auth_1 = auth_handler_cls
    query_1 = query_handler_cls
    query_2 = extended_query_handler_cls
    socket_timeout_1 = socket_timeout

    if not handler:
        class Handler(PsqlHandler):
            auth_handler_cls = auth_1
            query_handler_cls = query_1
            extended_query_handler_cls = query_2
            socket_timeout = socket_timeout_1

            def init_ssl(self, context: Context):
                context.key_file = key_file
                context.cert_file = cert_file

        handler = Handler

    with ThreadedTCPServer((host, port), handler) as server:
        # Activate the server; this will keep running until you
        # interrupt the program with Ctrl-C
        print(f"Running postgresql server on {host}:{port}")
        server.serve_forever()


def main():
    class TestAuth(AuthHandler):

        def handle(self, session: Session) -> Tuple[bool, Optional[str]]:
            if session.password == "test":
                return True, None

            return False, None

    class TestQueryHandler(QueryHandler):

        def handle(self, session: Session, sql) -> Tuple[Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:
            import pandas as pd
            import numpy as np
            from decimal import Decimal
            print(f"Handling query {sql}")
            return pd.DataFrame({'id': [1.232323, 2.1233, np.nan], 'name': [1, 2, 3],
                                 'p': [Decimal(1212323213123), Decimal(121312312321321323121321312321312323213),
                                       Decimal(123123213213213213213)]}), None

    import pandas as pd
    import numpy as np
    data_frame = pd.DataFrame({'id': [1.232323, 2.1233, np.nan], 'name': [1, 2, 3],
                               'p': [Decimal(1212323213123), Decimal(121312312321321323121321312321312323213),
                                     Decimal(123123213213213213213)]})

    class TestExtendedQueryHandler(ExtendedQueryHandler):

        def parse(self, session: Session, msg: ParseMessage) -> Tuple[
            Optional[Union[Any, Optional[PreparedStatement]]], Optional[Any]]:
            return PreparedStatement(
                name="test",
                sql=msg.query,
                parameters=[Param(oid_type=1, index=0, type=None), Param(oid_type=1, index=1, type=None)]
            ), None

        def bind(self, session: Session, prepared_statement: PreparedStatement, bind: BindMessage) -> Tuple[
            Optional[Union[Any, Optional[Portal]]], Optional[Any]]:
            return Portal(
                name=bind.portal_name,
                prepared_statement=prepared_statement,
                execution=None,
                row_description=data_frame,
            ), None

        def execute(self, session: Session, portal: Portal, max_rows: int) -> Tuple[
            Optional[Union[Any, Optional[DataFrame]]], Optional[Any]]:
            return data_frame, None

    run(auth_handler_cls=TestAuth, query_handler_cls=TestQueryHandler,
        extended_query_handler_cls=TestExtendedQueryHandler,
        key_file="tests/resources/server.key", cert_file="tests/resources/server.crt")


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Optional, List, Any

from pandas import DataFrame


@dataclass
class Param:
    oid_type: Optional[int]
    index: Optional[int]
    type: Optional[Any]

    @staticmethod
    def create_empty():
        return Param(oid_type=0, index=0, type=None)


@dataclass
class PreparedStatement:
    name: str
    sql: str
    parameters: Optional[List[Param]]


@dataclass
class Execution:
    max_rows: int
    offset: int
    df: Optional[DataFrame]


@dataclass
class Portal:
    name: str
    prepared_statement: PreparedStatement

    execution: Optional[Execution]

    # RowDescription message describing the rows that will be returned by executing the portal;
    row_description: Optional[DataFrame]



import struct
import io
import time
from typing import Callable, Any, List, Union, Optional, Tuple

from pypsql_api.wire.errors.core import NoData

packer_s1 = struct.Struct('b')
packer_s2be = struct.Struct('>h')
packer_s4be = struct.Struct('>i')
packer_s8be = struct.Struct('>q')

MAX_SIZE = 1024 * 1024 * 1024


def check_max_size(l):
    if l > MAX_SIZE:
        raise Exception(f"The message cannot be larger than {MAX_SIZE} bytes")


class ReadingIO:

    def __init__(self, buff: io.BufferedIOBase):
        self.buff = buff

    def read_bytes(self, n) -> bytes:
        data = self.buff.read(n)
        if not data:
            raise NoData(n, 0, f"No data. Expected {n} bytes")
        if len(data) != n:
            raise NoData(n, len(data), f"Not enough data. Expected {n} bytes got {len(data)}")

        return data

    def read_byte(self) -> bytes:
        return self.buff.read(1)

    def read_int32(self) -> int:
        data = self.read_bytes(4)
        return packer_s4be.unpack(data)[0]

    def read_int16(self) -> int:
        data = self.read_bytes(2)
        return packer_s2be.unpack(data)[0]

    def read_array_int16_head(self, fn: Callable[[int, Any], Any]) -> List[Any]:
        l = self.read_int16()
        return [fn(i, self) for i in range(l)]

    def read_array_int32_head(self, fn: Callable[[int, Any], Any]) -> List[Any]:
        l = self.read_int32()
        return [fn(i, self) for i in range(l)]

    def read_cstring(self, encoding='UTF-8'):
        s = io.BytesIO()
        b = self.read_byte()

        while b != b'\x00' and b != b'':
            s.write(b)

            b = self.read_byte()

        return s.getvalue().decode(encoding)

    def read_int32_delim_message(self, timeout=5) -> Tuple[Optional[Any], int]:
        """
        Reads the len = int32,
        then a bytes message of message_len = len - 4
        :return: a ReadingIO object or None
        """
        try:
            l = self.read_int32()
            msg_len = l - 4

            if msg_len <= 0:
                return None, 0

            check_max_size(msg_len)

            msg = self.read_bytes(msg_len)
            return ReadingIO(io.BytesIO(msg)), msg_len
        except NoData as e:
            if timeout < 1:
                raise e

            if e.got_bytes == 0:
                # idle till timeout on the connection
                time.sleep(1)
                return self.read_int32_delim_message(self, timeout=timeout - 1)


class WritingIO:

    def __init__(self, buff: io.BufferedIOBase):
        self.buff = buff

    def write_bytes(self, bts: Union[bytes, bytearray]):
        # BufferedIOBase will throw an exception if the bts cannot be written
        self.buff.write(bts)

    def write_byte(self, b: int):
        # BufferedIOBase will throw an exception if the bts cannot be written
        self.write_bytes(packer_s1.pack(b)[0:1])

    def write_int32(self, i: int):
        self.write_bytes(packer_s4be.pack(i))

    def write_int64(self, i: int):
        self.write_bytes(packer_s8be.pack(i))

    def write_int16(self, i: int):
        self.write_bytes(packer_s2be.pack(i))

    def write_array_int16_head(self, n: int, fn: Callable[[int, Any], Any]) -> List[Any]:
        self.write_int16(n)
        if n > 0:
            for i in range(n):
                fn(i, self)

    def write_array_int32_head(self, n: int, fn: Callable[[int, Any], Any]) -> List[Any]:
        self.write_int32(n)
        if n > 0:
            for i in range(n):
                fn(i, self)

    def write_cstring(self, v: str, encoding='UTF-8'):
        if v:
            self.write_bytes(
                v.encode(encoding)
            )
        self.write_bytes(b'\x00')

    def flush(self):
        if hasattr(self.buff, 'flush'):
            self.buff.flush()

    def close(self):
        self.buff.close()



from pandas import DataFrame

from pypsql_api.wire.actions_types import Names
from pypsql_api.context import Context


def process_simple_query(context: Context):
    msg = context.mem.get('message', None)

    if msg.query:
        msg.query = msg.query.replace('OPERATOR(pg_catalog.~)', '=') \
            .replace('~', '=')

        query_handler = context.query_handler
        resp, _ = query_handler.handle(session=context.session, sql=msg.query)

        if isinstance(resp, DataFrame):
            return context.update_mem('data', resp), Names.WRITE_DATA_FRAME

    return context, Names.WRITE_EMPTY_RESPONSE
